/******************************************************************************
 *
 * This file is provided under a dual license.  When you use or
 * distribute this software, you may choose to be licensed under
 * version 2 of the GNU General Public License ("GPLv2 License")
 * or BSD License.
 *
 * GPLv2 License
 *
 * Copyright(C) 2019 MediaTek Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See http://www.gnu.org/licenses/gpl-2.0.html for more details.
 *
 * BSD LICENSE
 *
 * Copyright(C) 2019 MediaTek Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *****************************************************************************/
/******************************************************************************
* [File]			msdc_basic.c
* [Version]			v1.0
* [Revision Date]	2011-05-04
* [Author]			haibing.xu@mediatek.inc, 20160329
* [Description]
*	MSDC Driver Source File
* [Copyright]
*	Copyright (C) 2011 MediaTek Incorporation. All Rights Reserved.
******************************************************************************/
#include "x_ldr_env.h"
#include "x_typedef.h"
#include "x_printf.h"
#include "x_hal_arm.h"
#include "x_assert.h"
#include "x_timer.h"
#include "x_os.h"
#include "x_bim.h"

#include "msdc_basic.h"
#include "msdc_host_hw.h"
#include "msdc_slave_hw.h"
#include "msdc_drv.h"
#include "msdc_mmc_protocal.h"

/*Privite data*/
extern 	int ch;
msdc_gpd_t DMA_MSDC_Header, DMA_MSDC_End;
msdc_bd_t DMA_BD[MAX_BD_PER_GPD];
UINT32 BDNum = 0;

//only recored the error information
status_error sts_err[] = {
		{MSDC_ADDRESS_OUT_OF_RANGE, "address out of raenge"},
		{MSDC_ADDRESS_MISALIGN, "address misalign"},
		{MSDC_BLOCK_LEN_ERROR, "block len error"},
		{MSDC_ERASE_SEQ_ERROR, "erase sequence error"},
		{MSDC_ERASE_PARAM_ERROR, "erase param error"},
		{MSDC_WP_VIOLATION, "access write protect error"},
		{MSDC_DEVICE_IS_LOCKED, "device is locked"},
		{MSDC_LOCK_UNLOCK_FAILED, "device lock unlock error"},
		{MSDC_COM_CRC_ERROR, "crc check error"},
		{MSDC_ILLEGAL_COMMAND, "illegal command"},
		{MSDC_DEVICE_ECC_FAILED, "device ecc failed"},
		{MSDC_CC_ERROR, "device met error"},
		{MSDC_GENERIC_ERROR, "a generic error"},
		{MSDC_CID_CSD_OVERWRITE, "cid/csd overwritten error"},
		{MSDC_WP_ERASE_SKIP, "erase skip error"},
		{MSDC_ERASE_RESET, "erase reset error"},
		{MSDC_SWITCH_ERROR, "switch error"},
		{MSDC_EXCEPTION_EVENT, "exception error"},
};

//Total 31 stages
UINT8 u1DrvLvlMap[32][2] =
							{
							{0,0},//level 0
							{1,0},
							{2,0},
							{3,0},
							{4,0},
							{5,0},
							{6,0},
							{7,0},
							{0,1},
							{1,1},
							{2,1},
							{3,1},
							{4,1}, //12
							{5,1},
							{6,1},
							{7,1},
							{0,2},
							{1,2},
							{2,2}, //18
							{3,2},
							{4,2},
							{5,2},
							{6,2},
							{7,2},
							{0,3},
							{1,3},
							{2,3},
							{3,3},
							{4,3},
							{5,3},
							{6,3},
							{7,3}//31
						};

#if defined(CC_MSDC_ENABLE) || defined(CC_SDMMC_SUPPORT)

VOID MsdcInfoInit(VOID)
{
	memset(&g_msdc_pub_info, 0 , sizeof(g_msdc_info));
	g_msdc_pub_info.RCA = MMC_DEF_RCA;
	g_msdc_pub_info.p_hs200_tuning = NULL;
	g_msdc_pub_info.p_hs400_tuning = NULL;
	g_msdc_pub_info.p_sd_tuning = NULL;
}

VOID MsdcStatusErrChk(INT32 status)
{
	INT32 i = 0;
	//1st, report error
	for (i=0; i < (sizeof(sts_err)/sizeof(status_error)); i++) {
		if (sts_err[i].error_bit & status)
			MSDC_LOG(MSG_LVL_ERR,"%s \n",sts_err[i].error_str);
	}
}

//assume use DMA to send cmd12/13
INT32 MsdcGetStatus(UINT32 *status)
{
	msdc_cmd cur_cmd;
	memset(&cur_cmd, 0, sizeof(msdc_cmd));

	if ((1 == ch) &&
		(g_msdc_pub_info.RCA  != MMC_DEF_RCA)		
		)
	{
		MSDC_LOG(MSG_LVL_ERR, "%s the RCA Not expected CH = %d! MsdcCurRCA = %d, Force use MMC %d \n",
				__FUNCTION__,ch, g_msdc_pub_info.RCA,MMC_DEF_RCA);
		g_msdc_pub_info.RCA = MMC_DEF_RCA;
	}

	cur_cmd.cmd_idx = CMD13_SEND_STATUS;
	cur_cmd.cmd_arg= (g_msdc_pub_info.RCA << 16);
	cur_cmd.flags |= DATA_MODE_BASIC_DMA;

	if (MSDC_SUCCESS == MsdcRequest(&cur_cmd)) {
		*status =  cur_cmd.resp[0];
		MsdcStatusErrChk(*status);
		return MSDC_SUCCESS;
	} else {
		return MSDC_FAILED;
	}
}

INT32 MsdcReqDataStop(VOID)
{
	msdc_cmd cur_cmd;
	memset(&cur_cmd, 0, sizeof(msdc_cmd));
    cur_cmd.cmd_idx = CMD12_STOP_TRANSMISSION;

	if (MSDC_SUCCESS == MsdcRequest(&cur_cmd))
		return MSDC_SUCCESS;
	else
		return MSDC_FAILED;
}

VOID MsdcStopDMA(VOID)
{
    MSDC_LOG(MSG_LVL_INFO, "DMA status: 0x%.8x\n",MSDC_READ32(DMA_CFG));
    MSDC_SETBIT(DMA_CTRL, DMA_CTRL_STOP);
    while(MSDC_READ32(DMA_CFG) & DMA_CFG_DMA_STATUS);
}

INT32 MsdcWaitHostIdle(UINT32 timeout)
{
	UINT32 i;

    for (i = 0; i < timeout * 1000; i++) {
        if ((0 == (MSDC_READ32(SDC_STS) & (SDC_STS_SDCBUSY | SDC_STS_CMDBUSY))) &&
			(0x00  == MSDC_READ32(MSDC_INT)))
		   return CMD_ERR_SUCCESS;

        HAL_Delay_us(1);
    }

	MSDC_LOG(MSG_LVL_ERR, "%s timeout Failed!\n",__FUNCTION__);
	return CMD_ERR_FAILED;
}

INT32 MsdcWaitClkStable (VOID)
{
    UINT32 i;

    for (i = 0; i < MSDC_CLK_TIMEOUT_LIMIT_COUNT * 1000; i++) {
        if (0 != (MSDC_READ32(MSDC_CFG) & MSDC_CFG_CARD_CK_STABLE))
            return MSDC_SUCCESS;

        HAL_Delay_us(1);
    }

	MSDC_LOG(MSG_LVL_ERR, "%s timeout Failed!\n",__FUNCTION__);
	return CMD_ERR_FAILED;
}

INT32 MsdcReset(VOID)
{
    UINT32 i;

    // Reset MSDC
    MSDC_SETBIT(MSDC_CFG, MSDC_CFG_RST);

    for (i = 0; i < MSDC_RST_TIMEOUT_LIMIT_COUNT; i++) {
        if (0 == (MSDC_READ32(MSDC_CFG) & MSDC_CFG_RST))
            return MSDC_SUCCESS;

        HAL_Delay_us(1000);
    }

	MSDC_LOG(MSG_LVL_ERR, "%s timeout Failed!\n",__FUNCTION__);
	return CMD_ERR_FAILED;
}

INT32 MsdcClrFifo(VOID)
{
    UINT32 i;

    // Reset FIFO
    MSDC_SETBIT(MSDC_FIFOCS, MSDC_FIFOCS_FIFOCLR);

    for (i = 0; i < MSDC_RST_TIMEOUT_LIMIT_COUNT * 1000; i++) {
        if (0 == (MSDC_READ32(MSDC_FIFOCS) & (MSDC_FIFOCS_FIFOCLR | MSDC_FIFOCS_TXFIFOCNT_MASK | MSDC_FIFOCS_RXFIFOCNT_MASK)))
            return MSDC_SUCCESS;

        HAL_Delay_us(1);
    }

	MSDC_LOG(MSG_LVL_ERR, "%s timeout Failed!\n",__FUNCTION__);
	return CMD_ERR_FAILED;
}

VOID MsdcChkFifo(VOID)
{
    // Check if rx/tx fifo is zero
    if ((MSDC_READ32(MSDC_FIFOCS) & (MSDC_FIFOCS_TXFIFOCNT_MASK | MSDC_FIFOCS_RXFIFOCNT_MASK)) != 0)
    {
        MSDC_LOG(MSG_LVL_INFO, " FiFo not 0, FIFOCS(0x%08X):0x%08X!!\n", MSDC_FIFOCS, MSDC_READ32(MSDC_FIFOCS));
        MsdcClrFifo();
    }
}

VOID MsdcClrIntr(VOID)
{
    // Check MSDC Interrupt vector register
    if (0 != MSDC_READ32(MSDC_INT))
    {
        MSDC_LOG(MSG_LVL_INFO, "MSDC INT(0x%08X) not 0:0x%08X!!\n", MSDC_INT, MSDC_READ32(MSDC_INT));

        // Clear MSDC Interrupt vector register
        MSDC_WRITE32(MSDC_INT, MSDC_READ32(MSDC_INT));
    }
}

INT32 MsdcErrorHandling(msdc_cmd *pCmd)
{
    INT32 i4Ret = MSDC_SUCCESS;
	UINT32 status = 0;

    // Reset MSDC
    MsdcReset();

    // Stop DMA
    if ((DATA_MODE_BASIC_DMA & pCmd->flags) || (DATA_MODE_DESC_DMA & pCmd->flags) || (DATA_MODE_ENHANCED_DMA & pCmd->flags))
        MsdcStopDMA();

    // Clear FIFO and wait it becomes 0
    i4Ret = MsdcClrFifo();
    if (MSDC_SUCCESS != i4Ret)
        goto ErrorEnd;

    // Clear MSDC interrupts and make sure all are cleared
    MsdcClrIntr();
    if (0x00  != MSDC_READ32(MSDC_INT)) {
        i4Ret = MSDC_FAILED;
        goto ErrorEnd;
    }
/*check status and send stop accordingly, also need wait the state */
#if 0
    // Send Stop Command for Multi-Write/Read
    if ((pCmd->pBuff) && ((pCmd->blocks * pCmd->block_size) > SDHC_BLK_SIZE)) {
    	MSDC_LOG(MSG_LVL_ERR, "Error handle, send stop for multi-block! \n");
        if (MsdcReqDataStop())
            MSDC_LOG(MSG_LVL_ERR, "MSDC fail to send stop cmd\n");
    } else
#endif
	{
		//check emmc status
		INT32 i4Chk_Tmo = 0;
		do
		{
			if (MSDC_SUCCESS != MsdcGetStatus(&status)) {
				MSDC_LOG(MSG_LVL_ERR, "%s Failed to get card status 0x%x \n", __FUNCTION__,status);
				goto ErrorEnd;
			}
			MSDC_LOG(MSG_LVL_INFO, "card status 0x%x in error handle \n",R1_CURRENT_STATE(status));

			if (R1_CURRENT_STATE(status) == R1_STATE_DATA ||
	    	    R1_CURRENT_STATE(status) == R1_STATE_RCV) {
				MSDC_LOG(MSG_LVL_ERR, "Error handle, send stop again in card status %d  \n",R1_CURRENT_STATE(status));
				if (MsdcReqDataStop()) {
	                MSDC_LOG(MSG_LVL_ERR, "MSDC fail to send stop cmd at line %d \n",__LINE__);
					//not care CMD12 stop fail
					//goto ErrorEnd;
	            }
	        } else if (R1_CURRENT_STATE(status) == R1_STATE_PRG) {
				; /*busy wait*/
			} else if (R1_CURRENT_STATE(status) != R1_STATE_TRAN) {
				MSDC_LOG(MSG_LVL_ERR, "ERROR! %s MMC state unexpected at line %d \n",__FUNCTION__,__LINE__);
				goto ErrorEnd;
			}

			//check timeout
			HAL_Delay_us(1);//1ms
			i4Chk_Tmo++;
			if (i4Chk_Tmo >= MSDC_STS_CHK_TIMEOUT) {
				MSDC_LOG(MSG_LVL_ERR, "ERROR! MSDC check card status timeout at line %d \n",__LINE__);
				goto ErrorEnd;
			}
		}while (!(status & MSDC_READY_FOR_DATA) || (R1_CURRENT_STATE(status) == R1_STATE_PRG));
    }
	MSDC_LOG(MSG_LVL_ERR, "Card Status: 0x%x\n", R1_CURRENT_STATE(status));

ErrorEnd:
    return i4Ret;
}

VOID MsdcFifoPathSelAndInit(VOID)
{

	#if MSDC_NEW_FIFO_PATH
	MSDC_LOG(MSG_LVL_ERR, "---MSDC Set FIFO path and Init---\n");
	MSDC_SET_FIELD(MSDC_IOCON, MSDC_IOCON_R_D_SMPL_SEL, 0);/*IOCON[5]*/
	MSDC_SET_FIELD(MSDC_IOCON, MSDC_IOCON_W_D_SMPL_SEL_SEL, 0);/*IOCON[9]*/
	MSDC_SET_FIELD(MSDC_IOCON, MSDC_IOCON_W_D_SMPL_SEL, 0);/*IOCON[8]*/

	MSDC_SET_FIELD(MSDC_IOCON, MSDC_IOCON_D_DLYLINE_SEL, 0);/*IOCON[3],clK tune all data Line share dly */

	MSDC_SET_FIELD(PAD_TUNE, PAD_RXDLY_SEL, 0);/*PAD_TUNE0[15] ,tune data path, tune mode select*/

	MSDC_SET_FIELD(PATCH_BIT2, CFG_CRCSTS, 1);/*PATCH_BIT2[28],latch CRC Status ,select async fifo path*/

	MSDC_SET_FIELD(PATCH_BIT2, CFG_RESP, 0);/*PATCH_BIT2[15],latch CMD Response ,select async fifo path*/

	/*eMMC50 Function Mux*/
	MSDC_SET_FIELD(EMMC50_CFG0, CRC_STS_SEL, 0); /*EMMC50_CFG0[4],write path switch to emmc45*/

	MSDC_SET_FIELD(EMMC50_CFG0, CFG_CMD_RESP_SEL, 0);/*EMMC50_CFG0[9],response path switch to emmc45*/

	MSDC_SET_FIELD(EMMC50_CFG0, CRC_STS_EDGE, 0);/*EMMC50_CFG0[3],select rising eage at emmc50 hs400*/

	/*Common Setting Config*/
	MSDC_SET_FIELD(PATCH_BIT0, CKGEN_MSDC_DLY_SEL, 0);/*patch_bit0[10], ckgen*/

	MSDC_SET_FIELD(PATCH_BIT1, CMD_RSP_TA_CNTR, 1); /*patch_bit1[3], TA*/

	MSDC_SET_FIELD(PATCH_BIT1, WRDAT_CRCS_TA_CNTR, 1);/*patch_bit1[0], TA*/

	MSDC_SET_FIELD(PAD_TUNE, PAD_CLK_TXDLY, 0);/*PAD_TUNE0[27],tx clk dly fix to 0 for HQA res*/
	MSDC_SET_FIELD(PATCH_BIT1, MSDC_PB1_GET_BUSY_MARGIN, 1);
#if defined(CC_MT5886)
	MSDC_SET_FIELD(PATCH_BIT1, MSDC_PB1_GET_CRC_MARGIN, 0);
#else
	MSDC_SET_FIELD(PATCH_BIT1, MSDC_PB1_GET_CRC_MARGIN, 1);
#endif

	MSDC_SET_FIELD(PATCH_BIT2, MSDC_BIT2_CRCSTS_LATCH_SEL, 0);
	/* LATCH_TA_EN Config for CMD Path non_HS400 */
	MSDC_SET_FIELD(PATCH_BIT2, MSDC_BIT2_RESP_LATCH_ENSEL, 0);
	MSDC_SET_FIELD(PATCH_BIT0, INT_DAT_LATCH_CK_SEL, 0);

	//set delay to default value

	// CMD_PAD_RDLY:
	MSDC_SET_FIELD(PAD_TUNE, PAD_CMD_RXDLY, 0/*31*/);
	MSDC_SET_FIELD(PAD_TUNE1, PAD_TUNE1_CMDRDLY2, 0);
	MSDC_SET_FIELD(PAD_TUNE, PAD_CMD_RD_RX_DLY_SEL, 0);
	MSDC_SET_FIELD(PAD_TUNE1, PAD_TUNE1_CMDRRDLY2SEL, 0);

	// DAT_PAD_RDLY:
	MSDC_SET_FIELD(PAD_TUNE, PAD_DAT_RD_RXDLY, 0/*31*/);
	MSDC_SET_FIELD(PAD_TUNE1, PAD_TUNE1_DATRRDLY2, 0);
	MSDC_SET_FIELD(PAD_TUNE, PAD_DAT_RD_RXDLY_SEL, 0);
	MSDC_SET_FIELD(PAD_TUNE1, PAD_TUNE1_DATRRDLY2SEL, 0);
    #else
	MSDC_LOG(MSG_LVL_ERR, "Warning:Not Support FIFO path arch\n");
	#endif

}

VOID MsdcDumpRegister(VOID)
{
	UINT32 i = 0;
	UINT32 u4regs [] = {
		0x0,0x4,0x8,0xc,
		0x10,0x14,0x18,0x1c,
		0x30,0x34,0x38,0x3c,
		0x40,0x44,0x48,0x4c,
		0x50,0x54,0x58,0x5c,
		0x60,0x64,
		0x70,0x74,0x78,0x7c,
		0x80,0x84,0x88,0x8c,
		0x90,0x94,0x98,0x9c,
		0xa0,0xa4,0xa8,
		0xb0,0xb4,0xb8,
		0xf0,0xf4,0xf8,0xfc,
		0x100,0x104,0x110,0x114,
		0x118,
		0x180,0x184,0x188,0x18c,
		0x190,0x194,0x198,0x19c,
		0x208,0x20c,0x228};

	for (i=0; i < sizeof(u4regs)/sizeof(UINT32); i++)
		MSDC_LOG(MSG_LVL_ERR,"[0X%x] 0x%x \n",u4regs[i],MSDC_READ32(0xf006d000 + u4regs[i]));


	MSDC_LOG(MSG_LVL_ERR,"\n EMMC Clock Src Setting - 0x%08X: 0x%08X\n", MSDC_CLK_S_REG1, MSDC_READ32(MSDC_CLK_S_REG1));
	MSDC_LOG(MSG_LVL_ERR,"\n SD Clock Src Setting - 0x%08X: 0x%08X\n", MSDC_CLK_S_REG0, MSDC_READ32(MSDC_CLK_S_REG0));
}

/*
 * Descriptor DMA support
 */

unsigned char MsdcCheckSumCal(unsigned char *buf, UINT32 len)
{
    UINT32 i = 0, sum = 0;

    for (; i<len;i++)
        sum += *(buf + i);

    return (0xFF-(unsigned char)sum);
}

VOID MsdcDescriptorConfig(VOID *pBuff, UINT32 u4BufLen)
{
    UINT32 i, tmpBuflen/*, fgEOL*/;
    unsigned char *tmppBuff = (unsigned char *)pBuff;

    if (u4BufLen == 0)
        return;

    BDNum = u4BufLen / BD_MAX_LEN + ((u4BufLen % BD_MAX_LEN == 0)?0:1);
    tmpBuflen = u4BufLen;

    MSDC_LOG(MSG_LVL_INFO, "\n----BD NUM: %d----\n", BDNum);

    if (BDNum > MAX_BD_NUM) {
        MSDC_LOG(MSG_LVL_WARN, "BD Number exceeds MAX value(%d)!\n", MAX_BD_NUM);
        return;
    }

    // Initial the structures
    memset(&DMA_MSDC_Header, 0x0, sizeof(msdc_gpd_t));
    memset(&DMA_MSDC_End, 0x0, sizeof(msdc_gpd_t));
    memset(DMA_BD, 0x0, sizeof(msdc_bd_t)*MAX_BD_NUM);

    // Config the BD structure array
    for (i = 0; i < BDNum; i++) {
        if (i != BDNum - 1)
            DMA_BD[i].pNext = (VOID *)(&DMA_BD[i+1]);
        else
            DMA_BD[i].eol = 1;

        DMA_BD[i].pBuff = (VOID *)(tmppBuff);
        DMA_BD[i].buffLen = (tmpBuflen>BD_MAX_LEN)?(BD_MAX_LEN):(tmpBuflen);
        tmpBuflen -= DMA_BD[i].buffLen;
        tmppBuff += DMA_BD[i].buffLen;

        DMA_BD[i].chksum = 0;
        DMA_BD[i].chksum = MsdcCheckSumCal((unsigned char *)(DMA_BD+i), 16);
    }

    // Config the GPD structure
    DMA_MSDC_Header.hwo = 1;  /* hw will clear it */
    DMA_MSDC_Header.bdp = 1;
    DMA_MSDC_Header.pNext = (VOID *)(&DMA_MSDC_End);
    DMA_MSDC_Header.pBuff = (VOID *)(DMA_BD);
    DMA_MSDC_Header.chksum = 0;  /* need to clear first. */
    DMA_MSDC_Header.chksum = MsdcCheckSumCal((unsigned char *)(&DMA_MSDC_Header), 16);

    HalFlushInvalidateDCacheMultipleLine((UINT32)(&DMA_MSDC_Header), sizeof(msdc_gpd_t));
    HalFlushInvalidateDCacheMultipleLine((UINT32)(&DMA_MSDC_End), sizeof(msdc_gpd_t));
    HalFlushInvalidateDCacheMultipleLine((UINT32)(DMA_BD), sizeof(msdc_bd_t)*BDNum);

    // Config the DMA HW registers
    MSDC_SETBIT(DMA_CFG, DMA_CFG_CHKSUM);
    MSDC_WRITE32(DMA_SA, 0x0);

#if defined(CC_MT5891) || defined(CC_MT5886) || defined(CC_MT5863)
    MSDC_WRITE32(DMA_SA_HIGH4BIT, 0x0);
#endif
    MSDC_SETBIT(DMA_CTRL, DMA_CTRL_BST_64);
    MSDC_SETBIT(DMA_CTRL, DMA_CTRL_DESC_MODE);

    MSDC_WRITE32(DMA_SA, (UINT32)(&DMA_MSDC_Header));
}

VOID MsdcDrvSet(UINT32 CLK, UINT32 CMD, UINT32 DAT)
{
#if defined(CC_MT5891)  || defined(CC_MT5886) || defined(CC_MT5863)
	MSDC_LOG(MSG_LVL_ERR,"==>Set  driving MORE . CLK %d ,CMD %d, DAT %d \n",CLK,CMD,DAT);

	MSDC_CLRBIT(0xF000D488, 0xFFF << 18);
	MSDC_CLRBIT(0xF000D48C, 0x3FFFF << 0);
	MSDC_CLRBIT(0xF000D480, 0xFFFFF << 12);

	//CLOCK
	MSDC_SETBIT(0xF000D488, ((u1DrvLvlMap[ CLK ][0] & 0x7)<<18));
	MSDC_SETBIT(0xF000D480, (u1DrvLvlMap[ CLK ][1] & 0x3) << 12);

	//CMD
	MSDC_SETBIT(0xF000D488, ((u1DrvLvlMap[ CMD ][0] & 0x7)<<21));
	MSDC_SETBIT(0xF000D480, (u1DrvLvlMap[ CMD ][1] & 0x3) << 14);

	//DAT
	MSDC_SETBIT(0xF000D480, ((u1DrvLvlMap[ DAT ][1] & 0x3) << 16) | ((u1DrvLvlMap[ DAT ][1] & 0x3) << 18) |
				((u1DrvLvlMap[ DAT ][1] & 0x3) << 20) | ((u1DrvLvlMap[ DAT ][1] & 0x3) << 22) |
				((u1DrvLvlMap[ DAT ][1] & 0x3) << 24) | ((u1DrvLvlMap[ DAT ][1] & 0x3) << 26) |
				((u1DrvLvlMap[ DAT ][1] & 0x3) << 28) | ((u1DrvLvlMap[ DAT ][1] & 0x3) << 30));

	MSDC_SETBIT(0xF000D488, ((u1DrvLvlMap[ DAT ][0] & 0x7)<<24) | ((u1DrvLvlMap[ DAT ][0] & 0x7)<<27));


	MSDC_SETBIT(0xF000D48C, ((u1DrvLvlMap[ DAT ][0] & 0x7) << 0) | ((u1DrvLvlMap[ DAT ][0]&0x7) << 3) |
			   ((u1DrvLvlMap[ DAT ][0] & 0x7) << 6) | ((u1DrvLvlMap[ DAT ][0] & 0x7) << 9) | ((u1DrvLvlMap[ DAT ][0]&0x7) << 12) | ((u1DrvLvlMap[ DAT ][0]&0x7) << 15));

#else
	MSDC_LOG(MSG_LVL_ERR,"WARNING! Not Support driving set, please add your driving setting  %s: line %d \n",__FUNCTION__,__LINE__);
#endif

}

//Add for HS200/HS400
UINT32 MsdcDrvGet(UINT32 IDX)
{
	UINT32 status = 0;
#if defined(CC_MT5891)  || defined(CC_MT5886) || defined(CC_MT5863)
	if(IDX == 0) //CLOCK
	{
		status = ((MSDC_READ32(0xF000D488)>>18) & 0x7) +
			((MSDC_READ32(0xF000D480)>>12) & 0x3)*8;
	}
	else if(IDX == 1) //CMD
	{
		status = ((MSDC_READ32(0xF000D488)>>21) & 0x7) +
			((MSDC_READ32(0xF000D480)>>14) & 0x3)*8;
	}
	else if(IDX == 2) //DAT
	{
		status = ((MSDC_READ32(0xF000D488)>>24) & 0x7) +
			((MSDC_READ32(0xF000D480)>>16) & 0x3)*8;
	}

#else
	MSDC_LOG(MSG_LVL_ERR,"WARNING! Not Support driving set, please add your driving setting  %s: line %d \n",__FUNCTION__,__LINE__);
#endif
	return status;
}

#endif
